from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Operator
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbtasks.models import Task


class TaskDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title"),
                dp.Field(key="in_charge", label="In Charge"),
                dp.Field(key="due_date", label="Due at"),
                dp.Field(key="description", label="Description"),
                dp.Field(key="tags", label="Tags"),
            ],
            formatting=[
                dp.Formatting(
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=Task.Status.UNSCHEDULED),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=Task.Status.STARTED),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=Task.Status.COMPLETED),
                        ),
                    ],
                    column="status",
                )
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            label=Task.Status.UNSCHEDULED.label,
                            icon=WBColor.BLUE_LIGHT.value,
                            value=Task.Status.UNSCHEDULED.value,
                        ),
                        dp.LegendItem(
                            label=Task.Status.STARTED.label,
                            icon=WBColor.YELLOW.value,
                            value=Task.Status.STARTED.value,
                        ),
                        dp.LegendItem(
                            label=Task.Status.COMPLETED.label,
                            icon=WBColor.GREEN_LIGHT.value,
                            value=Task.Status.COMPLETED.value,
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["status", "status", "status"],
                ["title", "title", "title"],
                ["creation_date", "due_date", "."],
                ["starting_date", "completion_date", "."],
                ["requester", "in_charge", "tags"],
                ["description", "description", "description"],
                ["comment", "comment", "comment"],
            ]
        )



from rest_framework import filters
from wbcore.filters import DjangoFilterBackend
from wbcore.viewsets import ModelViewSet, RepresentationViewSet

from wbtasks.filters import TaskFilter
from wbtasks.models import Task
from wbtasks.serializers import TaskModelSerializer, TaskRepresentationSerializer
from wbtasks.viewsets import TaskButtonConfig, TaskDisplayConfig, TaskTitleConfig


class TaskRepresentationViewSet(RepresentationViewSet):
    serializer_class = TaskRepresentationSerializer
    queryset = Task.objects.all()

    search_fields = ("title",)


class TaskModelViewSet(ModelViewSet):
    IDENTIFIER = "wbtasks:task"

    filter_backends = (
        DjangoFilterBackend,
        filters.SearchFilter,
        filters.OrderingFilter,
    )

    ordering_fields = ["due_date", "creation_date", "starting_date", "completion_date"]
    ordering = ["-due_date"]
    search_fields = ["title", "requester__computed_str", "assigned_to__computed_str", "description"]

    serializer_class = TaskModelSerializer
    filterset_class = TaskFilter
    queryset = Task.objects.all()

    display_config_class = TaskDisplayConfig
    button_config_class = TaskButtonConfig
    title_config_class = TaskTitleConfig



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbtasks.viewsets import viewsets

router = WBCoreRouter()
router.register(r"task", viewsets.TaskModelViewSet, basename="task")
router.register(r"taskrepresentation", viewsets.TaskRepresentationViewSet, basename="taskrepresentation")

urlpatterns = [
    path("", include(router.urls)),
]



from wbcore.menus import ItemPermission, MenuItem

TASK_MENUITEM = MenuItem(
    label="Tasks",
    endpoint="wbtasks:task-list",
    add=MenuItem(label="Create Task", endpoint="wbtasks:task-list"),
    permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["wbtasks.view_task"]),
)



import pytest
from wbcore.test import GenerateTest, default_config

config = {}
for key, value in default_config.items():
    config[key] = list(filter(lambda x: x.__module__.startswith("wbtasks"), value))


@pytest.mark.django_db
@GenerateTest(config)
class TestProject:
    pass



import pytest
from wbcore.contrib.authentication.factories import AuthenticatedPersonFactory

from wbtasks.viewsets.viewsets import TaskModelViewSet


@pytest.mark.django_db
class TestSpecificFilters:
    def test_filter_in_charge(self, task_factory):
        person = AuthenticatedPersonFactory()
        person2 = AuthenticatedPersonFactory()
        task_factory(in_charge=person2)
        mvs = TaskModelViewSet()
        qs = mvs.get_serializer_class().Meta.model.objects.all()
        assert mvs.filterset_class().filter_in_charge(qs, "", None) == qs
        assert mvs.filterset_class().filter_in_charge(qs, "", person).count() == 0
        assert mvs.filterset_class().filter_in_charge(qs, "", person2).count() == 1

    def test_filter_participants(self, task_factory):
        person = AuthenticatedPersonFactory()
        person2 = AuthenticatedPersonFactory()
        person3 = AuthenticatedPersonFactory()
        task_factory(
            in_charge=person2,
            assigned_to=(
                person2,
                person3,
            ),
        )
        mvs = TaskModelViewSet()
        qs = mvs.get_serializer_class().Meta.model.objects.all()
        assert mvs.filterset_class().filter_participants(qs, "", None) == qs
        assert mvs.filterset_class().filter_participants(qs, "", [person]).count() == 0
        assert mvs.filterset_class().filter_participants(qs, "", [person2]).count() == 1
        assert mvs.filterset_class().filter_participants(qs, "", [person3]).count() == 1
        assert mvs.filterset_class().filter_participants(qs, "", [person2, person3]).count() == 1



from __future__ import absolute_import, unicode_literals

from datetime import timedelta

from celery import shared_task
from django.db.models import Q
from django.utils import timezone
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.workers import Queue

from wbtasks.models import Task

NOTIFY_DUE_TASKS_INTERVAL_MINUTES = 60


@shared_task(queue=Queue.DEFAULT.value)
def notify_due_tasks(now=None, notification=None):
    if not now:
        now = timezone.now()
    for task in Task.objects.filter(
        ~Q(status=Task.Status.COMPLETED)
        & Q(due_date__gt=now)
        & Q(due_date__lt=now + timedelta(minutes=NOTIFY_DUE_TASKS_INTERVAL_MINUTES))
    ):
        if (profile := task.in_charge) and (user := profile.user_account):
            send_notification(
                code="wbtasks.task.notify",
                title="Task is due",
                body=f"the task {task.title} is due at {task.due_date:%d.%m.%Y}",
                user=user,
                reverse_name="wbtasks:task-detail",
                reverse_args=[task.id],
            )
    return notification


# from restbench.celery import app as celery_app

# celery_app.conf.beat_schedule.update({
#     'Tasks: Notify Due tasks': {
#         'task': 'wbtasks.tasks.notify_due_tasks',
#         'schedule': NOTIFY_DUE_TASKS_INTERVAL_MINUTES*60
#     }
# })



from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Entry, Person

from wbtasks.models import Task


def get_current_user_id(field, request, view):
    return request.user.profile.id


class TaskFilter(wb_filters.FilterSet):
    in_charge = wb_filters.ModelChoiceFilter(
        label="Assigned to",
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
        filter_params={"only_internal_users": True},
        # initial=get_current_user_id,
        method="filter_in_charge",
    )

    participants = wb_filters.ModelMultipleChoiceFilter(
        label="Participants",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        # initial=get_current_user_id,
        method="filter_participants",
    )

    def filter_in_charge(self, queryset, name, value):
        if value:
            return queryset.filter(in_charge=value)
        return queryset

    def filter_participants(self, queryset, name, value):
        if value:
            return queryset.filter(assigned_to__in=value).distinct()
        return queryset

    class Meta:
        model = Task
        fields = {
            "starting_date": ["gte", "exact", "lte"],
            "completion_date": ["gte", "exact", "lte"],
            "creation_date": ["gte", "exact", "lte"],
            "due_date": ["gte", "exact", "lte"],
            "requester": ["exact"],
            # "in_charge": ["exact"],
            "assigned_to": ["exact"],
            "status": ["exact"],
            "priority": ["exact"],
        }



from wbcore.metadata.configs.titles import TitleViewConfig


class TaskTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "{{title}}"

    def get_list_title(self):
        return "Tasks"

    def get_create_title(self):
        return "Create Task"



from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class TaskButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        return {
            bt.WidgetButton(key="see_activity", label="Linked Activity", icon=WBIcon.CALENDAR.icon),
            bt.WidgetButton(key="widget", label="Linked Widget", icon=WBIcon.LINK.icon),
        }



from datetime import date, timedelta

from django.utils.translation import gettext_lazy as _
from wbcore import serializers as wb_serializers
from wbcore import shares
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import (
    InternalUserProfileRepresentationSerializer,
    PersonRepresentationSerializer,
)
from wbcore.contrib.tags.serializers import TagSerializerMixin
from wbcore.metadata.configs.display.instance_display import (
    create_simple_section,
    repeat_field,
)

from wbtasks.models import Task


class TaskRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = Task
        fields = ("id", "title")


class TaskModelSerializer(TagSerializerMixin, wb_serializers.ModelSerializer):
    _requester = PersonRepresentationSerializer(source="requester")
    _in_charge = InternalUserProfileRepresentationSerializer(source="in_charge")
    _assigned_to = PersonRepresentationSerializer(source="assigned_to", many=True)

    @wb_serializers.register_resource()
    def ressource(self, instance, request, user):
        ressource = {}
        if instance.widget_endpoint:
            ressource["widget"] = instance.widget_endpoint
        return ressource

    def create(self, validated_data):
        if request := self.context.get("request"):
            validated_data["requester"] = request.user.profile
        return super().create(validated_data)

    def validate(self, attrs):
        if self.instance:
            assigned_to = attrs.get("assigned_to", self.instance.assigned_to)
            in_charge = attrs.get("in_charge", self.instance.in_charge)
        else:
            assigned_to = attrs.get("assigned_to", [])
            in_charge = attrs.get("in_charge", None)

        if assigned_to or in_charge:
            list_of_ids = []
            if in_charge:
                list_of_ids.append(in_charge.id)
            if assigned_to:
                if isinstance(assigned_to, list):
                    for person in assigned_to:
                        list_of_ids.append(person.id)
                else:
                    list_of_ids += assigned_to.values_list("id", flat=True)

            attrs["entities"] = Person.objects.filter(id__in=list_of_ids).distinct()

        return super().validate(attrs)

    class Meta:
        model = Task
        required_fields = ("title", "requester")
        read_only_fields = (
            "starting_date",
            "completion_date",
            "creation_date",
        )
        fields = (
            "id",
            "title",
            "starting_date",
            "completion_date",
            "creation_date",
            "due_date",
            "requester",
            "_requester",
            "in_charge",
            "_in_charge",
            "assigned_to",
            "_assigned_to",
            "status",
            "description",
            "comment",
            "priority",
            "tags",
            "_tags",
            "_additional_resources",
        )


class TaskToActivitySerializer(TaskModelSerializer):
    title = wb_serializers.CharField(label="Title", required=False)
    starting_date = wb_serializers.DateTimeField(label="Start")
    completion_date = wb_serializers.DateTimeField(label="End")
    requester = wb_serializers.PrimaryKeyRelatedField(queryset=Person.objects.all(), label="Creator")
    in_charge = wb_serializers.PrimaryKeyRelatedField(queryset=Person.objects.all(), label="In charge to")
    description = wb_serializers.TextField(label="Description")
    comment = wb_serializers.TextField(label="Review")

    class Meta:
        model = Task
        fields = (
            "id",
            "title",
            "starting_date",
            "completion_date",
            "requester",
            "_requester",
            "in_charge",
            "_in_charge",
            "assigned_to",
            "_assigned_to",
            "description",
            "comment",
            "_requester",
        )


@shares.register(
    section=create_simple_section(
        "task_section",
        _("Share as Task"),
        [
            [repeat_field(4, "task_share")],
            [repeat_field(4, "task_title")],
            [repeat_field(2, "task_due_date"), repeat_field(2, "task_in_charge")],
            [repeat_field(4, "task_description")],
        ],
        collapsed=True,
    )
)
class TaskShareSerializer(wb_serializers.Serializer):
    task_share = wb_serializers.BooleanField(default=False, label=_("Share as Task"))
    task_title = wb_serializers.CharField(
        max_length=256, label=_("Title"), required=False, depends_on=[{"field": "task_share", "options": {}}]
    )
    task_description = wb_serializers.TextField(
        label=_("Description"), required=False, depends_on=[{"field": "task_share", "options": {}}]
    )
    task_due_date = wb_serializers.DateField(
        label=_("Date"),
        required=False,
        default=lambda: date.today() + timedelta(days=7),
        depends_on=[{"field": "task_share", "options": {}}],
    )
    task_in_charge = wb_serializers.PrimaryKeyRelatedField(
        queryset=Person.objects.all(),
        label=_("In Charge"),
        required=False,
        depends_on=[{"field": "task_share", "options": {}}],
    )
    _task_in_charge = PersonRepresentationSerializer(
        source="task_in_charge", depends_on=[{"field": "task_share", "options": {}}]
    )
